//
// File:        UserOptions.java
// Package:     gov.llnl.babel
// Revision:     @(#) $Id: UserOptions.java 5340 2006-04-04 21:35:32Z epperly $
// Description: singleton Map for configuring general, user-related options
// 
// Copyright (c) 2000-2004, The Regents of the University of Calfornia.
// Produced at the Lawrence Livermore National Laboratory.
// Written by the Components Team <components@llnl.gov>
// UCRL-CODE-2002-054
// All rights reserved.
// 
// This file is part of Babel. For more information, see
// http://www.llnl.gov/CASC/components/. Please read the COPYRIGHT file
// for Our Notice and the LICENSE file for the GNU Lesser General Public
// License.
// 
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License (as published by
// the Free Software Foundation) version 2.1 dated February 1999.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the IMPLIED WARRANTY OF
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the terms and
// conditions of the GNU Lesser General Public License for more details.
// 
// You should have recieved a copy of the GNU Lesser General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

package gov.llnl.babel;

import gov.llnl.babel.backend.CodeGenerationException;
import gov.llnl.babel.backend.CodeGenerationFactory;
import gov.llnl.babel.backend.CodeGenerator;
import gov.llnl.babel.cli.CommandLineDictionary;
import gov.llnl.babel.cli.CommandLineExtension;
import gov.llnl.babel.cli.CommandLineSwitch;
import gov.llnl.babel.cli.InvalidArgumentException;
import gov.llnl.babel.cli.InvalidOptionException;
import gov.llnl.babel.cli.NameCollisionException;
import gov.llnl.babel.symbols.RegexMatch;
import gov.llnl.babel.symbols.RegexUnsupportedException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.regex.PatternSyntaxException;

/**
 * <code>UserOptions</code> is a singleton class that manages the
 * wide variety of settings available to a user.  The configuration 
 * options are read in through an array of strings (args).  Newly
 * specified arguments always replace the values of the previously 
 * specified ones with the exception of the repository path where
 * new arguments are appended to the existing list.
 *
 * For developers, to add a new internal Babel option. Create an inner
 * subclass in UserOptions with a name ending in <code>Switch</code>.
 * This class uses reflection to automatically add all inner classes
 * as command line switches.
 */

public class UserOptions implements CommandLineExtension
{ 

  private int d_numRequired = 0;
  private boolean d_versionSet = false;

  private static abstract class BasicSwitch implements CommandLineSwitch
  {
    private int d_shortForm;
    private String d_longName;
    private String d_helpText;
    public BasicSwitch(int shortForm, String longName, String helpText) {
      d_shortForm = shortForm;
      d_longName = longName;
      d_helpText = helpText;
    }
    
    final public String getLongName() { return d_longName; }
    final public int    getShortForm() { return d_shortForm; }
    final public void   setShortForm(int v) { d_shortForm = v; }
    final public String getHelpText() { return d_helpText; }
    public String getArgumentName() { return null; }
    public boolean hasRequiredArgument() { return false; }
    public boolean hasOptionalArgument() { return false; }
    public boolean isHidden() { return false; }
  }

  public class ExcludeExternalSwitch extends BasicSwitch {
    public ExcludeExternalSwitch() { 
      super('E', "exclude-external",
            "Code is generated only for the symbol hierarchies rooted at the symbols specified on the command line."); }
    final public void processCommandSwitch(String optarg) {
      System.err.println("Babel: Note: -E/--exclude-external is now the default behavior.");
      BabelConfiguration.getInstance().setExcludeExternal(true);
    }
    final public boolean isHidden() { return true; }
  }

  public class IncludeReferencedSwitch extends BasicSwitch {
    public IncludeReferencedSwitch() { 
      super(0, "include-referenced",
            "Code is generated for symbols referenced by those listed on the command line."); }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setExcludeExternal(false);
    }
  }

  public class GenerateHooksSwitch extends BasicSwitch {
    public GenerateHooksSwitch() { 
      super('i', "generate-hooks", "Generate pre-/post-method hooks"); 
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setGenerateHooks(true);
    }
  }

  public class LanguageSubdirSwitch extends BasicSwitch {
    public LanguageSubdirSwitch() { 
      super('l', "language-subdirs",
            "Generate code in a language-specific subdirectory.");
    }
    final public void processCommandSwitch(String optarg) 
    {
      BabelConfiguration.getInstance().setMakeLanguageSubdir(true);
    }
  }

  public class LanguageSubdirOffSwitch extends BasicSwitch {
    public LanguageSubdirOffSwitch() { 
      super(0, "language-subdirs-off",
            "Turn off --language-subdirs and restore default behavior.");
    }
    final public void processCommandSwitch(String optarg) 
    {
      BabelConfiguration.getInstance().setMakeLanguageSubdir(false);
    }
  }

  public class GenerateSubdirSwitch extends BasicSwitch {
    public GenerateSubdirSwitch() { 
      super('g', "generate-subdirs",
            "Postpend Java-style package subdirs."); 
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setMakePackageSubdirs(true);
    }
  }

  public class GenerateSubdirOffSwitch extends BasicSwitch {
    public GenerateSubdirOffSwitch() { 
      super(0, "generate-subdirs-off",
            "Turn off --generate-subdirs and restore default behavior."); 
    }
    final public void processCommandSwitch(String optarg) 
    {
      BabelConfiguration.getInstance().setMakePackageSubdirs(false);
    }
  }

  public class ParseCheckSwitch extends BasicSwitch {
    public ParseCheckSwitch() { 
      super('p', "parse-check",
            "Parse the sidl file but do not generate code.");
    }
    final public void processCommandSwitch(String optarg) 
      throws InvalidOptionException
    {
      BabelConfiguration.getInstance().setParseCheckOnly(true);
      if (BabelConfiguration.getMultiMode()) {
        throw new InvalidOptionException("--parse-check cannot be used after --multi");
      }
      if (d_numRequired++ > 0) {
        throw new InvalidOptionException("Only one of the following options or its equivalent is allowed: --help, --version, --client, --server, --text or --parse-check unless you use the --multi option.");
      }
    }
  }

  public class ExperimentalParserSwitch extends BasicSwitch {
    public ExperimentalParserSwitch() { 
      super('P', "old-parser",
            "Parse the sidl file with old, deprecated parser.");
    }
    final public void processCommandSwitch(String optarg) 
      throws InvalidOptionException
    {
      BabelConfiguration.getInstance().setExperimentalParser(false);
      System.err.println("babel: warning using old, deprecated parser.");
      if (BabelConfiguration.getMultiMode()) {
        throw new InvalidOptionException("--old-parser cannot be used after --multi");
      }
    }
  }
  
  public class VerboseSwitch extends BasicSwitch {
    public VerboseSwitch() { 
      super(0, "verbose",
            "Turn on verbose debugging text");
    }
    final public void processCommandSwitch(String optarg) 
      throws InvalidOptionException
    {
      BabelConfiguration.getInstance().setVerbose(true);
      if (BabelConfiguration.getMultiMode()) {
        throw new InvalidOptionException("--verbose cannot be used after --multi");
      }
    }
  }
  
  public class HideGlueSwitch extends BasicSwitch {
    public HideGlueSwitch() { 
      super('u', "hide-glue",
            "Put \"glue\" (not modifiable by user) generate code in a glue/ subdirectory.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setMakeGlueSubdirs(true);
    }
  }

  public class HideGlueOffSwitch extends BasicSwitch {
    public HideGlueOffSwitch() { 
      super(0, "hide-glue-off",
            "Turn off --hide-glue and restore default behavior.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setMakeGlueSubdirs(false);
    }
  }

  public class VersionSwitch extends BasicSwitch {
    public VersionSwitch() { 
      super('v', "version",
            "Display the version information and exit.");
    }
    final public void processCommandSwitch(String optarg) 
      throws InvalidOptionException
    {
      ++d_numRequired;
      d_versionSet = true;
      System.out.println( "Babel version " + Version.VERSION );
    }
  }

  public class CxxIORExceptionSwitch extends BasicSwitch {
    public CxxIORExceptionSwitch() { 
      super('x', "cxx-ior-exception",
            "Include IOR pointer checking in C++ stubs; throw a NullIORException if IOR is Null.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setCxxCheckNullIOR(true);
    }
  }

  public class GenerateSidlStdlibSwitch extends BasicSwitch {
    public GenerateSidlStdlibSwitch() { 
      super(0, "generate-sidl-stdlib",
            "Regenerate only the SIDL standard library.");
    }

    final public void processCommandSwitch(String optarg) 
      throws InvalidOptionException
    {
      BabelConfiguration.getInstance().setGenerateStdlib(true);
    }
  }

  public class NoDefaultRepositorySwitch extends BasicSwitch {
    public NoDefaultRepositorySwitch() { 
      super(0, "no-default-repository",
            "Prohibit use of default repository to resolve symbols.");
    }
    final public void processCommandSwitch(String optarg) {
      System.err.println("Babel: Warning: no-default-repository is a babel driver script option only.");
    }
  }

  public class SuppressTimestampSwitch extends BasicSwitch {
    public SuppressTimestampSwitch() { 
      super(0, "suppress-timestamp",
            "Suppress timestamps in generated files.");
    }
    final public void processCommandSwitch(String optarg) {
      System.err.println("Babel: Note: --suppress-timestamp is now the default behavior.");
      BabelConfiguration.getInstance().setSuppressTimestamps(true);
    }
    final public boolean isHidden() { return true; }
  }

  public class TimestampSwitch extends BasicSwitch {
    public TimestampSwitch() { 
      super(0, "timestamp",
            "Generate timestamps in generated files.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setSuppressTimestamps(false);
    }
  }

  public class CommentLocalOnlySwitch extends BasicSwitch {
    public CommentLocalOnlySwitch() { 
      super(0, "comment-local-only",
            "Only comment locally defined methods in stub.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setCommentLocalOnly(true);
    }
  }

  public class ShortFileNamesSwitch extends BasicSwitch {
    public ShortFileNamesSwitch() { 
      super(0, "short-file-names",
            "Use in conjuction with -g to generate file names that don't include package names.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setShortFileNames(true);
    }
  }

  public class SuppressIORSwitch extends BasicSwitch {
    public SuppressIORSwitch() { 
      super(0, "suppress-ior",
            "Refrain from generating IOR source or header files.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.getInstance().setSuppressIOR(true);
    }
  }

  public class MultiModeSwitch extends BasicSwitch {
    public MultiModeSwitch() { 
      super(0, "multi",
            "Enable generation of multiple targets.");
    }
    final public void processCommandSwitch(String optarg) {
      BabelConfiguration.setMultiMode(true);
    }
  }

  private static abstract class RequiredArgSwitch extends BasicSwitch 
  {
    private String d_argName;
    public RequiredArgSwitch(int shortForm, String longName, String helpText,
                             String argName){
      super(shortForm, longName, helpText);
      d_argName = argName;
    }
    public boolean hasRequiredArgument() { return true; }
    public String getArgumentName() { return d_argName; }
  }

  public class AssertionLevelSwitch extends RequiredArgSwitch
  {
    public AssertionLevelSwitch() {
      super('a', "assertion-level", 
            "Generate SIDL-specified assertion checking code. NOTE: 0=None, 1=Support, and 2=Advanced.", "lvl");
    }

    /**
     * Check if the specified assertion checking level is supported.
     */
    private boolean validAssertionLevel( String alvl ) { 
      return  BabelConfiguration.NO_CHECKING.equals(alvl) ||
        BabelConfiguration.SUPPORT_CHECKING.equals(alvl) ||
        BabelConfiguration.ADVANCED_CHECKING.equals(alvl);
    }

    final public void processCommandSwitch(String optarg) 
      throws InvalidArgumentException
    {
      if ((optarg != null) && validAssertionLevel(optarg)) {
        BabelConfiguration.getInstance().setAssertionLevel(optarg);
      }
      else {
        throw new InvalidArgumentException
          ("The assertion level must be 0, 1, or 2.");
      }
    }
  }

  public class ClientSwitch extends RequiredArgSwitch
  {
    public ClientSwitch() {
      super('c', "client", 
            "Generate only client code in the specified language (C | C++ | UC++ | F77 | F90 | Java | Python).", "lang");
    }
    
    /**
     * Check if the specified client-side language is supported.
     */
    private String validClientLanguage( String lang ) 
      throws InvalidArgumentException
    { 
      CodeGenerationFactory factory = CodeGenerationFactory.getInstance();
      CodeGenerator cg = factory.getCodeGenerator(lang, "stub");
      if (null != cg) {
        try {
          cg.setName(lang);
        }
        catch (CodeGenerationException cge) {
          throw new InvalidArgumentException(cge.getMessage());
        }
        return cg.getName();
      }
      return null;
    }
    
    final public void processCommandSwitch(String optarg) 
      throws InvalidArgumentException, InvalidOptionException
    {
      optarg = optarg.toLowerCase();
      optarg = validClientLanguage(optarg);
      if (null != optarg) {
        BabelConfiguration.getInstance().setGenerateClient(true);
        BabelConfiguration.getInstance().setTargetLanguage(optarg);
        if ((d_numRequired++ > 0) && !BabelConfiguration.getMultiMode()) {
          throw new InvalidOptionException("Only one of the following options or its equivalent is allowed: --help, --version, --client, --server, --text, or --parse-check unless you use the --multi option.");
        }
      }
      else {
        throw new InvalidArgumentException
          ("Unsupported language specified for client option.");
      }
    }
  }

  public class ExcludeSwitch extends RequiredArgSwitch
  {
    public ExcludeSwitch() {
      super('e', "exclude", 
            "Symbols matching the regular expression are excluded from code generation.", "regex");
    }

    final public void processCommandSwitch(String optarg) 
      throws InvalidArgumentException
    {
      try {
        BabelConfiguration.getInstance().addExcluded(new RegexMatch(optarg));
      }
      catch(RegexUnsupportedException rue) {
        throw new InvalidArgumentException
          ("The --exclude feature requires a Java runtime environment with the 1.4 library or higher for regular expression support.");
      }
      catch(PatternSyntaxException PSE) { 
        throw new InvalidArgumentException
          ("The regular expression is invalid: " + PSE.getMessage());
      }
    }
  }

  public class MakePrefixSwitch extends RequiredArgSwitch
  {
    public MakePrefixSwitch() {
      super('m', "make-prefix",
            "<prefix> is prepended to the name of babel.make and the symbols defined internally to allow Babel to be run multiple times in a single directory.", "prefix");
    }

    final public void processCommandSwitch(String optarg) 
    {
      BabelConfiguration.getInstance().setMakePrefix(optarg);
    }
  }

  public class OutputDirectorySwitch extends RequiredArgSwitch
  {
    public OutputDirectorySwitch() {
      super('o', "output-directory",
            "Set Babel output directory (\'.\' default).", "dir");
    }

    final public void processCommandSwitch(String optarg) 
    {
      BabelConfiguration.getInstance().setOutputDirectory(optarg);
    }
  }

  public class RepositoryPathSwitch extends RequiredArgSwitch
  {
    public RepositoryPathSwitch() {
      super('R', "repository-path",
            "Set semicolon-separated URL list used to resolve symbols.", 
            "path");
    }

    final public void processCommandSwitch(String optarg) 
      throws InvalidArgumentException
    {
      String shortname = optarg;
      boolean absolutePath = false;
      java.io.File[] roots = java.io.File.listRoots();

      //directory ends with sperator? 
      if(!shortname.endsWith(""+java.io.File.separatorChar)) 
        shortname = shortname + java.io.File.separatorChar;
      //Figure out if we have an absolute path by checking that the path
      //begins with a root.  ( '/' in Unix, "C:\" or "D:\" in Windows)
      for(int i = 0; i < roots.length; ++i) {
        if (shortname.startsWith(roots[i].toString())) {
          absolutePath=true;
          break;
        }
      }
      
      if(absolutePath) { //If it's an absolute path, good
        BabelConfiguration.getInstance().addToRepositoryPath(shortname);
      } else {// if the path is realitive, make it absolute
        java.io.File file = new java.io.File(".");
        try {
          BabelConfiguration.getInstance().addToRepositoryPath(file.getCanonicalPath()
                                        + java.io.File.separatorChar + shortname);
        } catch (java.io.IOException e) {
          throw new InvalidArgumentException
            ("Unable to get path to current directory! Any XML output may contain an incorrect path!");
        }
      }
    }
  }

  public class ServerSwitch extends RequiredArgSwitch
  {
    public ServerSwitch() {
      super('s', "server", 
            "Generate server (and client code) in the specified language (C | DC++ | UC++ | F77 | F90 | Java | Python).", "lang");
    }

    /**
     * Check if the specified client-side language is supported.
     */
    private String validServerLanguage( String lang ) 
      throws InvalidArgumentException
    { 
      CodeGenerationFactory factory = CodeGenerationFactory.getInstance();
      CodeGenerator cg = factory.getCodeGenerator(lang, "skel");
      if (null != cg) {
        try {
          cg.setName(lang);
        }
        catch (CodeGenerationException cge) {
          throw new InvalidArgumentException(cge.getMessage());
        }
        return cg.getName();
      }
      return null;
    }

    final public void processCommandSwitch(String optarg) 
      throws InvalidArgumentException, InvalidOptionException
    {
      optarg = optarg.toLowerCase();
      optarg = validServerLanguage(optarg);
      if (null != optarg) {
        BabelConfiguration.getInstance().setGenerateServer(true);
        BabelConfiguration.getInstance().setTargetLanguage(optarg);
        if (d_numRequired++ > 0 && !BabelConfiguration.getMultiMode()) {
          throw new InvalidOptionException("Only one of the following options or its equivalent is allowed: --help, --version, --client, --server, --text, or --parse-check unless you use the --multi option.");
        }
      }
      else {
        throw new InvalidArgumentException
          ("Unsupported language specified for server option.");
      }
    }
  }

  public class TextSwitch extends RequiredArgSwitch
  {
    public TextSwitch() {
      super('t', "text", 
            "Generate text in specified form (XML | SIDL | HTML).", "form");
    }

    /**
     * Check if the specified client-side language is supported.
     */
    private String validTextForm( String lang ) 
      throws InvalidArgumentException
    { 
      CodeGenerationFactory factory = CodeGenerationFactory.getInstance();
      CodeGenerator cg = factory.getCodeGenerator(lang, "text");
      if (null != cg) {
        try {
          cg.setName(lang);
        }
        catch (CodeGenerationException cge) {
          throw new InvalidArgumentException(cge.getMessage());
        }
        return cg.getName();
      }
      return null;
    }

    final public void processCommandSwitch(String optarg) 
      throws InvalidArgumentException, InvalidOptionException
    {
      optarg = optarg.toLowerCase();
      optarg = validTextForm(optarg); 
      if (null != optarg) {
        BabelConfiguration.getInstance().setGenerateText(true);
        BabelConfiguration.getInstance().setTargetLanguage(optarg);
        if (d_numRequired++ > 0 && !BabelConfiguration.getMultiMode()) {
          throw new InvalidOptionException("Only one of the following options or its equivalent is allowed: --help, --version, --client, --server, --text, or --parse-check unless you use the --multi option.");
        }
      }
      else {
        throw new InvalidArgumentException
          ("Unsupported text form specified for text option.");
      }
    }
  }

  public class VpathSwitch extends RequiredArgSwitch
  {
    public VpathSwitch() {
      super('V', "vpath", 
            "Prepend alternative search path for reading Impl splicer blocks. Does not affect where Impls are generated. NOTE: --vpath=. is a no-op for autoconf/automake. If you really want current directory, use another argument like `pwd` or even ./.", "path");
    }

    final public void processCommandSwitch(String optarg) 
    {
      BabelConfiguration.getInstance().setVPathDirectory(optarg);
    }
  }

  /**
   * The handle to singleton instance of this class.
   */
  private static UserOptions s_my_instance = null;
  
  /**
   * Since this is a singleton class, the constructor is protected.
   */
  protected UserOptions() {
  }

  public void registerCommandLineSwitches(CommandLineDictionary dict)
    throws NameCollisionException
  {
    Class myClass = getClass();
    Class[] myClasses= myClass.getClasses();
    Class[] paramList = new Class[1];
    Object[] args = new Object[1];
    paramList[0] = myClass;
    args[0] = this;
    int i;
    for(i = 0; i < myClasses.length; ++i) {
      Class innerClass = myClasses[i];
      if (innerClass.getName().endsWith("Switch") &&
          !innerClass.isInterface()) {
        try {
          Constructor ctor = innerClass.getConstructor(paramList);
          Object o = ctor.newInstance(args);
          if (o instanceof CommandLineSwitch) {
            dict.addCommandLineSwitch((CommandLineSwitch)o);
          }
        }
        catch (NoSuchMethodException nsme) {
          // ignore
        }
        catch (InvocationTargetException ite) {
          System.err.println("InvocationTargetException: " + ite.getMessage());
        }
        catch (IllegalArgumentException iarg) {
          System.err.println("IllegalArgumentException: " + iarg.getMessage());
        }
        catch (InstantiationException ie) {
          System.err.println("InstantiationException: " + ie.getMessage());
        }
        catch (IllegalAccessException iae) {
          System.err.println("IllegalAccessException: " + iae.getMessage());
        }
      }
    }
  }

  /**
   * Return the singleton instance of the user options.  
   * If the user options instance has not yet been created, 
   * then it will be created by this call.
   */
  public static UserOptions getInstance() {
    if (s_my_instance == null) {
      s_my_instance = new UserOptions();
    }
    return s_my_instance;
  }

  public void reset()
  {
    d_numRequired = 0;
    d_versionSet = false;
  }

  /**
   * Return the number of required elements on the command line.
   */
  public int getNumRequired() { return d_numRequired; }

  /**
   * Return <code>true</code> iff the version option appeared and it
   * was printed.
   */
  public boolean getVersionPrinted() { return d_versionSet; }
}
